import abc
import cmath

from hilbert import EQ_ROUND_TO

from hilbert import stock

import numpy


class PolarComplex(stock.Repr):
    def __init__(self, norm, phase):
        self.norm, self.phase = norm.real, phase.real

    def __str__(self):
        return f'{self.norm}·ej{self.phase}'

    def __eq__(self, other):
        return not self.__ne__(other)

    def __ne__(self, other):
        return bool(round(abs(self.number - (
            other.number if isinstance(other, self.__class__) else other)
        ), EQ_ROUND_TO))

    @property
    def number(self):
        return cmath.rect(self.norm, self.phase)

    @classmethod
    def from_x(cls, *numbers):
        arr = numpy.array([cls(*cmath.polar(number)) for number in numbers])

        return arr[0] if arr.shape[0] == 1 else arr

    @staticmethod
    def ph(*phases):
        return (sum(map(lambda x: x.real, phases)) - cmath.pi) % (2*cmath.pi) - cmath.pi

    def eat(self, other):
        if isinstance(other, (int, float, complex)):
            return self.__class__(*cmath.polar(other))

        if isinstance(other, self.__class__):
            return other

        raise NotImplementedError(f'Operation with {other}')

    def eatbin(self, other, norm_phase):
        pc = self.eat(other)

        return self.__class__(*norm_phase(pc.norm, pc.phase))

    def take(self, other):
        if isinstance(other, (int, float, complex)):
            return other

        if isinstance(other, self.__class__):
            return other.number

        raise NotImplementedError(f'Operation with {other}')

    def takebin(self, other, norm_phase):
        number = self.take(other)

        return self.__class__(*norm_phase(number))

    def conjugate(self):
        return self.__class__(self.norm, -self.phase)

    def __bool__(self):
        return bool(self.norm)

    def __pos__(self):
        return self

    def __neg__(self):
        return self.__class__(self.norm, self.ph(self.phase, cmath.pi))

    def __abs__(self):
        return self.norm

    def __round__(self, to=4):
        norm_to, phase_to = (to, to) if isinstance(to, int) else to

        return self.__class__(round(self.norm, norm_to), round(self.phase, phase_to))

    def __mul__(self, other):
        return self.eatbin(other, lambda on, op: (self.norm*on, self.ph(self.phase, op)))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self.eatbin(other, lambda on, op: (self.norm/on, self.ph(self.phase, -op)))

    def __rtruediv__(self, other):
        return self.eatbin(other, lambda on, op: (on/self.norm, self.ph(op, -self.phase)))

    def __rpow__(self, other):
        n = self.number
        return self.eatbin(other, lambda on, op: (
            cmath.exp(-op*n.imag)*on**n.real, n.imag*cmath.log(on) + op*n.real))

    def __pow__(self, other):
        return self.takebin(other, lambda n: (
            cmath.exp(-self.phase*n.imag)*self.norm**n.real,
            n.imag*cmath.log(self.norm) + self.phase*n.real))

    def __add__(self, other):
        return self.takebin(other, lambda num: cmath.polar(self.number + num))

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self.takebin(other, lambda num: cmath.polar(self.number - num))

    def __rsub__(self, other):
        return -self.__sub__(other)


class Scale(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def num_prod(self, number):
        """Return the product"""

    def __mul__(self, number):
        if not number:
            return 0
        elif number == 1:
            return self
        elif isinstance(number, (int, float, complex)):
            return self.num_prod(number)
        else:
            raise NotImplementedError(f'Product by {number}')

    def __rmul__(self, number):
        return self.__mul__(number)

    def __pos__(self):
        return self

    def __neg__(self):
        return self.__mul__(-1)

    def __truediv__(self, number):
        return self.__mul__(1/number)

    def __rtruediv__(self, number):
        raise NotImplementedError(f'{self} is not /-invertible')


class AbelianSumScale(Scale, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def eat(self, other):
        """Return equivalent of `other` - if any"""

    @abc.abstractmethod
    def add_other(self, other):
        """Add other instance"""

    def __add__(self, other):
        if not other:
            return self

        other = self.eat(other)

        if isinstance(other, self.__class__):
            return self.add_other(other)

        raise NotImplementedError(f'Addition to {repr(other)}')

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self.__add__(-other)

    def __rsub__(self, other):
        return -self.__sub__(other)


class Vector(AbelianSumScale, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def braket(self, other):
        """Real scalar product"""

    def __matmul__(self, other):
        if not other:
            return 0

        other = self.eat(other)

        if isinstance(other, self.__class__):
            return self.braket(other)

        raise NotImplementedError(f'Braket with {repr(other)}')

    def __rmatmul__(self, other):
        return numpy.conj(self.__matmul__(other))



"""Curve library"""
import abc

import numpy
from numpy.polynomial import polynomial

from hilbert import EQ_ROUND_TO

from hilbert import algebra
from hilbert import stock

EXPONENTS = {0: '', 1: '', 2: '²', 3: '³', 4: '⁴', 5: '⁵', 6: '⁶', 7: '⁷', 8: '⁸', 9: '⁹'}


class InvalidParameters(Exception):
    """Raised when curve input parameters do not suit the curve type"""


# Abstract base curves #########################################################

@stock.FrozenLazyAttrs(('parameters', 'pole'))
class Curve(stock.Repr, stock.Hashable, algebra.Scale,
            metaclass=abc.ABCMeta):
    min_dof = None

    def __init__(self, *parameters, pole=0):
        self.parameters = self._clean_parameters(parameters)
        self.pole = pole

    def __call__(self, x: numpy.array):
        return self.evaluate(x - self.pole)

    def _clean_parameters(self, params):
        if self.min_dof and len(params) < self.min_dof:
            raise InvalidParameters(f'At least {self.min_dof} parameter(s) required')

        return params

    @abc.abstractmethod
    def evaluate(self, s: numpy.array):
        """Normalized function"""

    def eqkey(self):
        return self.parameters, self.pole

    def kind(self):
        return self.__class__.__name__

    def svar(self, exponent=1):
        if exponent == 0:
            return ''

        exp = EXPONENTS.get(abs(exponent), f'^({abs(exponent)})')

        return ('/' if exponent < 0 else '') + (
            f'(x - {self.pole})'.replace('- -', '+ ') if self.pole else 'x') + exp


class CurveFormatted(Curve):
    def __str__(self):
        return self.format(*self.parameters)

    @abc.abstractmethod
    def format(self, *params):
        """Text representation given parameters `params`"""


class LinearCurve(CurveFormatted, metaclass=abc.ABCMeta):
    def num_prod(self, number):
        return self.__class__(*(number*p for p in self.parameters), pole=self.pole)


class NonLinearCurve(CurveFormatted, metaclass=abc.ABCMeta):
    """Of the form c0*f(c1, ...) with parameters (c0, c1, ...)"""
    min_dof = 2

    @abc.abstractmethod
    def evaluate_normal(self, s: numpy.array):
        """Normalized function with c0 == 1"""

    def evaluate(self, s: numpy.array):
        return self.parameters[0]*self.evaluate_normal(s)

    def num_prod(self, number):
        return self.__class__(
            number*self.parameters[0], *self.parameters[1:], pole=self.pole)


class BasisCurve(Curve):
    def scale(self, k0, k):
        return self.__class__(
            k0*self.parameters[0], *self.parameters[1:], pole=self.pole*k)


# Piecewise curves #############################################################

@stock.FrozenLazyAttrs(('jumps_at', 'piece_count', 'curves'))
class PiecewiseCurve(Curve):
    def __init__(self, jumps_at, curves):
        super().__init__(pole=0)
        self.jumps_at, self.piece_count = tuple(jumps_at), len(jumps_at) + 1
        self.curves = curves

    def eqkey(self):
        return self.jumps_at, self.curves

    def num_prod(self, number):
        return self.__class__(self.jumps_at, tuple(c*number for c in self.curves))

    def __str__(self):
        return ' | '.join(list(map(str, self.curves)))

    def kind(self):
        headed = zip(self.jumps_at, self.curves[1:])
        chain = ''.join([self.curves[0].kind()] + [
            f'[{head}]{curve.kind()}' for head, curve in headed])

        return f'PW:{chain}'

    def evaluate(self, s: numpy.array):
        functions = (self.curves[b.nonzero()[0][0]] for b in self.conditions(s))

        return numpy.array([f(v) for f, v in zip(functions, s)])

    def conditions(self, s: numpy.array):
        return numpy.array([s < self.jumps_at[0]] + [
            self.jumps_at[i] <= s < self.jumps_at[i+1] for i in range(self.piece_count - 2)
        ] + [s >= self.jumps_at[self.piece_count-2]]).transpose()


# Non-linear curves ############################################################

class Exp(NonLinearCurve, BasisCurve):
    """Orthonormal Fourier basis vector with a first parameter 1/(√measure√dimension)"""

    def evaluate_normal(self, s: numpy.array):
        return numpy.exp(self.parameters[1]*s)

    def format(self, coef, exp_coef):
        return f'({coef})exp({exp_coef}){self.svar()}'


class Gaussian(NonLinearCurve):
    def evaluate_normal(self, s: numpy.array):
        return numpy.exp(self.parameters[1]*s**2 - 1j*self.parameters[2]*s)

    def format(self, coef, exp2_coef, exp1_coef):
        return f'({coef})exp' + (
            f'({exp2_coef}){self.svar(2)}' if not exp1_coef else
            f'[({exp2_coef}){self.svar(2)} - ({1j*exp1_coef}){self.svar(1)}]')


class XtoA(NonLinearCurve):
    def evaluate_normal(self, s: numpy.array):
        return s**self.parameters[1]

    def format(self, coef, exp_coef):
        return f'({coef}){self.svar(exp_coef)}'


# Linear curves ################################################################

class Delta(LinearCurve, BasisCurve):
    """Orthonormal basis vector with a first parameter 1/√measure"""

    def evaluate(self, s: numpy.array):
        return numpy.where(s == 0, self.parameters[0], 0)

    def format(self, inverse_sqrt_measure):
        return f'({inverse_sqrt_measure})δ{self.svar()}'


class Polynomial(LinearCurve):
    def evaluate(self, s: numpy.array):
        return polynomial.Polynomial(self.parameters)(s)

    def kind(self):
        return f'Poly({len(self.parameters) - 1})'

    def format(self, *params):
        return ' + '.join([f'({param}){self.svar(d)}' for d, param in enumerate(params)])


class Log(LinearCurve):
    def evaluate(self, s: numpy.array):
        return self.parameters[0]*numpy.log(s)

    def format(self, coefficient):
        return f'({coefficient})log{self.svar()}'


class Xlog(LinearCurve):
    def evaluate(self, s: numpy.array):
        return self.parameters[0]*s*numpy.log(s)

    def format(self, coefficient):
        return f'({coefficient}){self.svar()}log{self.svar()}'


class InverseXPolynomial(LinearCurve):
    def evaluate(self, s: numpy.array):
        return polynomial.Polynomial([0] + list(reversed(self.parameters)))(1/s)

    def kind(self):
        return f'Poly(-{len(self.parameters)})'

    def format(self, *params):
        return ' + '.join(reversed([f'({param}){self.svar(-n - 1)}'
                                    for n, param in enumerate(reversed(params))]))


@stock.FrozenLazyAttrs(('image',))
class ImageCurve(Curve):
    """General, non-analytical definition for vectors"""

    def __init__(self, image):
        super().__init__(pole=0)
        self.image = image

    def __str__(self):
        return str(self.image)

    def eqkey(self):
        return self.pole, tuple(self.image.i.round(EQ_ROUND_TO))

    def num_prod(self, number):
        return self.__class__(self.image.__class__(series=number*self.image.i))

    def evaluate(self, s: numpy.array):
        return numpy.array([self.image[x] for x in s])



import operator

import numpy

from hilbert import EQ_ROUND_TO

from hilbert import algebra
from hilbert import stock

from hilbert.curves import lib
from hilbert.curves import vectors


@stock.FrozenLazyAttrs(('space', 'o'))
class Operator(stock.Eq, stock.IndexXYFrame):
    def __init__(self, data_frame, space):
        super().__init__(data_frame)
        self.space = space

    def __bool__(self):
        return not self.where(self.o != 0).o.dropna(how='all').empty

    def eq(self, other):
        return self.space == other.space and not round(abs(self - other))

    def take(self, other):
        if not isinstance(other, self.__class__):
            return other

        if not other.space == self.space:
            raise NotImplementedError('Operators not based on the same space')

        return other.o

    def un(self, op):
        return self.__class__(op(self.o), self.space)

    def bin(self, other, op):
        return self.__class__(op(self.o, self.take(other)), self.space)

    def rbin(self, other, op):
        return self.__class__(op(self.take(other), self.o), self.space)

    def apply(self, function, *args, **kwargs):
        return self.un(lambda df: df.apply(function, *args, **kwargs))

    def rename(self, *args, **kwargs):
        return self.un(lambda df: df.rename(*args, inplace=False, **kwargs))

    def toggle_polar(self):
        if self.is_polar():
            return self.apply(lambda col: col.map(lambda pc: pc.number))

        return self.apply(lambda arr: algebra.PolarComplex.from_x(*arr), raw=True)

    def is_polar(self):
        return set(self.o.dtypes).pop() == numpy.object

    def where(self, condition, *args, **kwargs):
        return self.un(lambda df: df.where(condition, *args, **kwargs))

    def dagger(self):
        return self.un(lambda df: numpy.conj(df.transpose()))

    def __round__(self, to=EQ_ROUND_TO):
        if self.is_polar():
            return self.apply(lambda col: col.map(lambda x: round(x, to)))

        return self.apply(lambda col: col.round(to))

    def __pos__(self):
        return self.un(operator.pos)

    def __neg__(self):
        return self.un(operator.neg)

    def __abs__(self):
        return self.un(lambda df: df.abs())

    def __lt__(self, other):
        return self.bin(other, operator.lt).o

    def __gt__(self, other):
        return self.bin(other, operator.gt).o

    def __le__(self, other):
        return self.bin(other, operator.le).o

    def __ge__(self, other):
        return self.bin(other, operator.ge).o

    def __mul__(self, other):
        return self.bin(other, operator.mul)

    def __rmul__(self, other):
        return self.rbin(other, operator.mul)

    def __truediv__(self, other):
        return self.bin(other, operator.truediv)

    def __rtruediv__(self, other):
        return self.rbin(other, operator.truediv)

    def __add__(self, other):
        return self.bin(other, operator.add)

    def __radd__(self, other):
        return self.rbin(other, operator.add)

    def __sub__(self, other):
        return self.bin(other, operator.sub)

    def __rsub__(self, other):
        return self.rbin(other, operator.sub)

    def transform(self, vector):
        return vector.__class__(self.space, lib.ImageCurve(
            vector.image_type(series=self.o@vector.image.i)))

    def __matmul__(self, other):
        if isinstance(other, vectors.Vector):
            if not self.space == other.space:
                raise NotImplementedError('Operator and vector not based on the same space')

            return self.transform(other)

        return self.bin(other, operator.matmul)

    def __pow__(self, other):
        return self.bin(other, operator.pow)

    def is_hermitian(self):
        return self.dagger() == self

    def is_unitary(self):
        return self.dagger()@self == self.space.Id and self@self.dagger() == self.space.Id


class ROperator(stock.RealIndexMixin, Operator):
    """ℝ-indexed operator"""


class COperator(stock.ComplexIndexMixin, Operator):
    """ℂ-indexed operator"""



import abc

import numpy

from hilbert import spaces
from hilbert import stock


@stock.FrozenLazyAttrs(('space',))
class System(metaclass=abc.ABCMeta):
    def __init__(self, space):
        self.space = space

    @abc.abstractmethod
    def __call__(self, tf, ti):
        """Return time development operator from `ti` to `tf`"""

    def evolve(self, tf, ti, vector):
        return self(tf, ti)@vector

    def show_evolution(self, tf, ti, vector, ncurves, **kwargs):
        self.space.show_vectors(
            *(self.evolve(t, ti, vector) for t in numpy.linspace(ti, tf, ncurves)),
            **kwargs)

    def show_density_evolution(self, tf, ti, vector, ncurves, **kwargs):
        self.space.show_vectors_density(
            *(self.evolve(t, ti, vector) for t in numpy.linspace(ti, tf, ncurves)),
            **kwargs)

    def variance(self, operator, vector):
        return self.mean(operator@operator, vector) - self.mean(operator, vector)**2

    @staticmethod
    def mean(operator, vector):
        return (vector@(operator@vector))/(vector@vector)


@stock.FrozenLazyAttrs(lazy_keys=('hamiltonian',))
class HamiltonianSystem(System):
    def __call__(self, ti, tf):
        return self.space.unitary_op(-self.hamiltonian*(tf - ti), validate=False)


class R1System(System):
    def __init__(self, lbound, rbound, dimension):
        super().__init__(spaces.R1Field.range(
            spaces.LebesgueCurveSpace, lbound, rbound, dimension))

    @property
    def position_op(self):
        return self.space.operator(numpy.diag(self.space.bases.domain()))

    @property
    def momentum_op(self):
        Pp = self.space.operator(numpy.diag(self.space.fourier_labels))
        F = self.space.fourier_op

        return F@Pp@F.dagger()


class QuasiFreeParticleR1(R1System, HamiltonianSystem):
    def _make_hamiltonian(self):
        Hp = self.space.operator(numpy.diag(self.space.fourier_labels**2))
        F = self.space.fourier_op

        return F@Hp@F.dagger()



import abc
import functools

from matplotlib import pyplot


class Repr(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def __str__(self):
        """Object's text content"""

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self}>'


class Eq(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def eq(self, other):
        """Return self == other for same type"""

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False

        return self.eq(other)

    def __ne__(self, other):
        return not self.__eq__(other)


class Hashable(Eq, metaclass=abc.ABCMeta):
    def __hash__(self):
        return hash(self.eqkey())

    @abc.abstractmethod
    def eqkey(self):
        """Return hashable, frozen property to compare to others"""

    def eq(self, other):
        return self.eqkey() == other.eqkey()


class ImmutableReset(Exception):
    """Raised on reset try of immutable attributes"""


class FrozenLazyAttrs:
    def __init__(self, frozen_keys=(), lazy_keys=()):
        self.frozen_keys, self.lazy_keys = frozen_keys, lazy_keys

    def __call__(self, cls):
        self.init_keys(cls)
        cls.__init__ = self.decorate_init(cls.__init__)

        for key in self.frozen_keys:
            setattr(cls, key, property(
                functools.partial(self.get, key), functools.partial(self.set, key)))

        for key in self.lazy_keys:
            setattr(cls, key, property(
                functools.partial(self.setget, key), functools.partial(self.set, key)))

        return cls

    def init_keys(self, obj):
        self_frozen = dict.fromkeys(self.frozen_keys + self.lazy_keys, False)
        obj._frozen = {**getattr(obj, '_frozen', {}), **self_frozen}

    def decorate_init(self, function):
        def init(instance, *args, **kwargs):
            self.init_keys(instance)
            function(instance, *args, **kwargs)

        return init

    @staticmethod
    def set(key, instance, value):
        if instance._frozen[key] is True:
            raise ImmutableReset(
                f"Immutable attribute '{key}' already set to {getattr(instance, key)}")

        setattr(instance, f'_{key}', value)
        instance._frozen[key] = True

    @staticmethod
    def setget(key, instance):
        if instance._frozen[key] is False:
            value = getattr(instance, f'_make_{key}')()
            setattr(instance, f'_{key}', value)
            instance._frozen[key] = True

            return value

        return getattr(instance, f'_{key}')

    @staticmethod
    def get(key, instance):
        return getattr(instance, f'_{key}')


class IndexFrame(Repr, metaclass=abc.ABCMeta):
    """Pandas data frame wrapper for custom indexing with complex numbers"""
    def __init__(self, data_frame):
        self.o = data_frame

    def __str__(self):
        return f'\n{self.o}'

    @abc.abstractstaticmethod
    def to_index(x):
        """Map coordinate `x` to index label"""

    @abc.abstractstaticmethod
    def to_coordinate(label):
        """Map index label to coordinate"""

    def index_domain(self, index, copy=False):
        return index.map(self.to_coordinate).to_numpy(dtype=self.dtype, copy=copy)

    def domain(self, copy=False):
        return self.index_domain(self.o.index, copy)


class ComplexIndexMixin:
    @staticmethod
    def to_index(x):
        if x is not None:
            return x.real, x.imag

    @staticmethod
    def to_coordinate(label):
        return complex(*label)

    @staticmethod
    def multiindex_at(i, j, multiindex):
        return multiindex[len(multiindex.levels[1])*i + j]


class RealIndexMixin:
    @staticmethod
    def to_index(x):
        return x

    @staticmethod
    def to_coordinate(label):
        return label


class IndexXFrame(IndexFrame):
    def __getitem__(self, loc):
        return self.o.loc[self.location(*loc)]

    def __setitem__(self, loc, value):
        self.o.loc[self.location(*loc)] = value

    def location(self, xloc, column_loc):
        return ((slice(*map(self.to_index, (xloc.start, xloc.stop)))
                if isinstance(xloc, slice) else self.to_index(xloc)), column_loc)

    def at(self, x, column):
        return self.o.at[self.to_index(x), column]

    def setat(self, x, column, value):
        self.o.at[self.to_index(x), column] = value


class IndexXYFrame(IndexFrame):
    def __getitem__(self, loc):
        return self.o.loc[tuple(map(self.location, loc))]

    def __setitem__(self, loc, value):
        self.o.loc[tuple(map(self.location, loc))] = value

    def location(self, xloc):
        return (slice(*map(self.to_index, (xloc.start, xloc.stop)))
                if isinstance(xloc, slice) else self.to_index(xloc))

    def at(self, x, y):
        return self.o.at[self.to_index(x), self.to_index(y)]

    def setat(self, x, y, value):
        self.o.at[self.to_index(x), self.to_index(y)] = value


class PyplotShow:
    golden = 1.61803398875

    def __init__(self, **fig_kwds):
        self.fig_kwds = fig_kwds
        self.fig_kwds.setdefault('figsize', (2*self.golden*6, 6))
        self.fig_kwds.setdefault('facecolor', 'w')

    def __call__(self, plot_method):
        @functools.wraps(plot_method)
        def wrapper(obj, *args, fig_kwds=None, **kwargs):
            figure, axes = pyplot.subplots(**{**self.fig_kwds, **(fig_kwds or {})})
            plot_method(obj, *args, figure=figure, axes=axes, **kwargs)
            pyplot.show()

            return figure, axes

        return wrapper



import abc
import itertools

import pandas
import numpy

from hilbert import EQ_ROUND_TO, INDEX_ROUND_TO

from hilbert import algebra
from hilbert import operators
from hilbert import stock

from scipy import linalg

from hilbert.curves import lib
from hilbert.curves import vectors


class InvalidVectorMade(Exception):
    """Raised for wrong vector input"""


@stock.FrozenLazyAttrs(('bases',))
class Space(stock.Repr, metaclass=abc.ABCMeta):
    def __init__(self, bases, validate_basis=False):
        self.operator_type, self.image_type = (
            (operators.COperator, vectors.CImage) if isinstance(bases, C1Field) else
            (operators.ROperator, vectors.RImage) if isinstance(bases, R1Field) else
            (None, None))

        if self.operator_type is None:
            raise NotImplementedError('`bases` argument should be a `Field`')

        self.bases, self.validate_basis = bases, validate_basis

    def __str__(self):
        return repr(self.bases)

    @property
    def Id(self):
        return self.operator_type(pandas.DataFrame(
            numpy.identity(self.bases.dimension),
            columns=self.bases.o.index, index=self.bases.o.index), self)

    def __call__(self, *args, validate=True, **kwargs):
        vector = self.make_vector(*args, **kwargs)

        if validate:
            self.validate(vector)

        return vector

    def __getitem__(self, loc):
        """Lazy basis vector definition using `loc` indexing"""
        xloc, key = loc

        if not isinstance(key, str):
            raise NotImplementedError('A single basis should be specified')

        if key not in self.bases.o:
            self.bases.o[key] = numpy.nan

        vec = self.bases[xloc, key]

        if isinstance(vec, pandas.Series):
            nans_at = vec[vec.map(lambda v: not isinstance(v, algebra.Vector))].index

            if len(nans_at):
                def_method = self.try_def_method(key)

                for ix in nans_at:
                    self.bases.o.at[ix, key] = def_method(self.bases.to_coordinate(ix))

                return self.bases[xloc, key]
        elif not isinstance(vec, algebra.Vector):
            def_method = self.try_def_method(key)
            vector = def_method(xloc)
            self.bases.setat(xloc, key, vector)

            return vector

        return vec

    def try_def_method(self, key):
        def_method = getattr(self, f'def_{key}_vector', None)

        if def_method is None:
            raise KeyError(f"Requested new vector(s) for non-analytic basis '{key}'")

        return def_method

    def __setitem__(self, key, basis):
        """Set basis"""
        self.bases.o[key] = basis

    @abc.abstractmethod
    def make_vector(self, *args, **kwargs):
        """Vector instance constructor returning new vector"""

    @abc.abstractmethod
    def scale_basis_vector(self, k, vector):
        """Keep norm invariant"""

    def validate(self, vector):
        braket = vector @ vector

        if not (braket > 0 and numpy.isfinite(braket)):
            raise InvalidVectorMade(
                f'{vector} does not belong to the space - no finite norm!')

    def scale(self, k):
        self.bases.scale_cell(k)
        self.bases.scale_index(k)

        for key in self.bases.o:
            self[key] = self.bases.o[key].map(
                lambda vector: self.scale_basis_vector(k, vector)
                if isinstance(vector, algebra.Vector) else vector)

    def extend(self, copies=1):
        replicas = self.bases.replicas(copies)
        vector_replicas = {key: list(
            self.bases.yield_vector_replicas(key, copies, replicas))
            for key in filter(lambda key: not self.is_analytic(key), self.bases.o)}
        self.bases.extend_index(copies, replicas)
        self.bases.update()

        for key in set(self.bases.o) - set(vector_replicas):
            getattr(self, f'extend_{key}_basis')(copies)

        for key, repls in vector_replicas.items():
            for z, series in repls:
                self.bases.setat(z, key, self(series=series, validate=False))

    def operator(self, *args, **kwargs):
        return self.operator_type(pandas.DataFrame(
            *args, columns=self.bases.o.index, index=self.bases.o.index, **kwargs
        ), self)

    def op_from_callable(self, function, *apply_args, **apply_kwds):
        return self.operator().apply(function, *apply_args, **apply_kwds)

    def unitary_op(self, hermitian, validate=True):
        if validate and not hermitian.is_hermitian():
            raise NotImplementedError('Operator.unitary requires an hermitian generator')

        H = (hermitian if not hermitian.is_polar() else hermitian.toggle_polar())

        return self.operator(linalg.expm(1j*H.o.to_numpy()))

    @property
    def fourier_op(self):
        if not (self.operator_type == operators.ROperator and self.bases.dimension % 2 == 0):
            raise NotImplementedError('Fourier basis is defined over ℝ with even dimension')

        return self.op_from_callable(self.fourier_basis_coords, axis=1)

    @property
    def fourier_labels(self):
        return pandas.Series(self.fourier_label(self.bases.domain()),
                             index=self.bases.o.index, name='p')

    def fourier_label(self, x):
        return 2*numpy.pi*(
            (x - self.bases.bounds[0])/(self.bases.measure*self.bases.dimension) - 0.5
        )/self.bases.measure

    def fourier_basis_coords(self, row):
        return self.fourier_labels.map(
            lambda p: numpy.exp(-1j*p*row.name)/numpy.sqrt(self.bases.dimension))

    def map_basis(self, operator, new_key='new', key='delta'):
        if not operator.space == self:
            raise NotImplementedError('Operator should belong to the space')

        self[new_key] = self[:, key].map(lambda v: operator@v)

    def is_orthonormal(self, key):
        basis = self[:, key]
        brakets = self.op_from_callable(lambda c: basis.map(lambda v: basis[c.name]@v))

        return brakets == self.Id

    def is_basis(self, key):
        return all([u == self.vector(key, self.coords(key, u)) for u in self[:, 'delta']])

    def is_analytic(self, key):
        return hasattr(self, f'def_{key}_vector') and hasattr(self, f'extend_{key}_basis')

    def coords(self, key, vector):
        return self[:, key].map(lambda v: v@vector)

    def vector(self, key, coords):
        return sum(coord*vector for coord, vector in zip(coords, self[:, key]))

    @stock.PyplotShow(nrows=2, ncols=1)
    def show_basis_slice(self, key, from_x=None, to_x=None, **kwargs):
        top_ax, _ = self.plot_vectors(
            self[from_x:to_x, key], *kwargs.pop('axes'), **kwargs)

        if hasattr(self, f'{key}_labels'):
            labels = getattr(self, f'{key}_labels')
            fr = '' if from_x is None else labels[from_x]
            to = '' if to_x is None else labels[to_x]
            name = labels.name
        else:
            fr, to = '' if from_x is None else from_x, '' if to_x is None else to_x
            name = 'x'

        top_ax.set_title(f'{key} basis ∀{name} ∈ ({fr}...{to})')

    @stock.PyplotShow(nrows=2, ncols=1)
    def show_vectors(self, *vectors, **kwargs):
        self.plot_vectors(vectors, *kwargs.pop('axes'), **kwargs)

    @stock.PyplotShow()
    def show_vectors_density(self, *vectors, **kwargs):
        self.plot_vectors_density(vectors, kwargs.pop('axes'), **kwargs)

    @staticmethod
    def plot_vectors(vectors, top_ax, bottom_ax, **kwargs):
        for vector in vectors:
            vector.full_plot(top_ax, bottom_ax, **kwargs)

        top_ax.yaxis.set_label_text('Re')
        bottom_ax.yaxis.set_label_text('Im')
        top_ax.grid(), bottom_ax.grid()

        return top_ax, bottom_ax

    @staticmethod
    def plot_vectors_density(vectors, axes, **kwargs):
        for vector in vectors:
            vector.density_plot(axes, **kwargs)

        axes.yaxis.set_label_text('abs²')
        axes.grid()

        return axes


class LebesgueCurveSpace(Space):
    def def_delta_vector(self, x):
        return self(lib.Delta(1/numpy.sqrt(self.bases.measure), pole=x),
                    validate=self.validate_basis)

    def def_fourier_vector(self, x):
        return self(lib.Exp(1/numpy.sqrt(self.bases.dimension*self.bases.measure),
                            -1j*self.fourier_label(x)), validate=self.validate_basis)

    def make_vector(self, *args, series=None, **kwargs):
        if series is None:
            return vectors.Vector(self, *args, **kwargs)

        return vectors.Vector(self, lib.ImageCurve(self.image_type(
            series=series.reindex(index=self.bases.o.index, fill_value=0))))

    def scale_basis_vector(self, k, vector):
        return vector.__class__(self, *(
            c.scale(self.norm_scale_factor(k), k) for c in vector.curves))

    def norm_scale_factor(self, k):
        return {operators.ROperator: 1/numpy.sqrt(k),
                operators.COperator: 1/k}[self.operator_type]

    def extend_delta_basis(self, copies):
        """Nothing to do - ready to be extended on demand"""

    def extend_fourier_basis(self, copies):
        p_series = self.fourier_labels
        vectors = self.bases.o['fourier'].dropna().to_numpy()
        self['fourier'] = numpy.nan

        for vector in vectors:
            p = -vector.curves[0].parameters[1].imag
            new_x = p_series[(p_series - p).round(EQ_ROUND_TO) == 0].index[0]
            self.bases.setat(new_x, 'fourier', vector/numpy.sqrt(2*copies + 1))


class Field(stock.IndexXFrame, metaclass=abc.ABCMeta):
    dtype = None

    @property
    def dimension(self):
        return len(self.o.index)

    @abc.abstractproperty
    def measure(self):
        """The cell measure"""

    @abc.abstractproperty
    def bounds(self):
        """Return the boundary values as [min, max], ..."""

    @abc.abstractproperty
    def limits(self):
        """Return the 'open' limits as (l, r), ..."""

    @abc.abstractproperty
    def cell_limits(self):
        """Return cell boundary values as (b0, b1, ...), ... for each dimension"""

    @abc.abstractmethod
    def scale_index(self, k):
        """Scale the index by `k`"""

    @abc.abstractmethod
    def scale_cell(self, k):
        """Scale the unit cell length by `k`"""

    @abc.abstractmethod
    def replicas(self, copies):
        """Replicate the index from its boundaries - return array of replicas"""

    @abc.abstractmethod
    def yield_vector_replicas(self, key, copies, index_replicas):
        """Yield all new `z, series` tuples for basis `key`"""

    @abc.abstractmethod
    def extend_index(self, copies, index_replicas):
        """Reindex from the replicas"""

    @abc.abstractmethod
    def plot_domain(self, **kwargs):
        """Show space domain"""

    def update(self):
        self.o = self.o.apply(lambda s: s.map(
            lambda v: v.update() if isinstance(v, algebra.Vector) else v))


class R1Field(stock.RealIndexMixin, Field):
    dtype = numpy.float64

    def __init__(self, cell, *args, **kwargs):
        self.cell = cell
        super().__init__(pandas.DataFrame(*args, **kwargs))

    def __str__(self):
        return (f'ℝ-segment | cell {round(self.cell, 9)} with bounds {self.bounds}'
                f' | D = {self.dimension}')

    @property
    def measure(self):
        return self.cell

    @property
    def limits(self):
        mi, ma = self.bounds

        return (mi - self.cell/2, ma + self.cell/2)

    @property
    def cell_limits(self):
        return tuple(self.o.index - self.cell/2) + self.limits[1:]

    @property
    def bounds(self):
        return [self.o.index.min(), self.o.index.max()]

    def scale_index(self, k):
        self.o.index *= k

    def scale_cell(self, k):
        self.cell *= k

    def replicas(self, copies):
        l, r = self.bounds

        return [numpy.around(self.o.index + n*(r - l + self.measure), INDEX_ROUND_TO)
                for n in range(-copies, copies + 1)]

    def yield_vector_replicas(self, key, copies, index_replicas):
        current_vectors = self.o[key][~self.o[key].isna()]
        cases = itertools.product(enumerate(current_vectors), index_replicas)

        for (k, vec), index in cases:
            series = pandas.Series(vec.image.i.to_numpy(), index=index)

            yield index[k], series

    def extend_index(self, copies, index_replicas):
        self.o = self.o.reindex(
            index=self.o.index.__class__(itertools.chain(*index_replicas)))

    @stock.PyplotShow(figsize=(25, 1))
    def plot_domain(self, **kwargs):
        axes = self.base_plot(kwargs.pop('axes'))
        axes.grid(which='minor', color='w')
        axes.patch.set_facecolor('green')

    def base_plot(self, axes, title=None):
        axes.set_title(title or str(self))
        axes.xaxis.set_ticks(self.cell_limits, minor=True)
        axes.yaxis.set_ticks([])
        axes.set_xlim(*self.limits)
        axes.set_ylim(-self.cell/2, self.cell/2)
        axes.set_aspect('equal')
        axes.xaxis.set_label_text('Re')

        return axes

    @classmethod
    def range(cls, space_type, start, end, cells):
        arr, cell = numpy.linspace(start, end, cells, retstep=True)

        return space_type(cls(cell, index=pandas.Index(
            numpy.around(arr, INDEX_ROUND_TO)).astype(cls.dtype)))

    @classmethod
    def from_index(cls, space_type, index):
        return space_type(cls(
            index._step if isinstance(index, pandas.RangeIndex)
            else abs((index[1:] - index[:-1]).array).mean(), index=index))


class C1Field(stock.ComplexIndexMixin, Field):
    dtype = numpy.complex128

    def __init__(self, re_cell, im_cell, *args, **kwargs):
        self.re_cell, self.im_cell = re_cell, im_cell
        super().__init__(pandas.DataFrame(*args, **kwargs))

    def __str__(self):
        return (f'ℂ-rectangle | cell {self.re_cell}x{self.im_cell} with '
                f'bounds {self.bounds} | D = {self.dimension}')

    @property
    def measure(self):
        return self.re_cell*self.im_cell

    @property
    def limits(self):
        (rmi, rma), (imi, ima) = self.bounds

        return ((rmi - self.re_cell/2, rma + self.re_cell/2),
                (imi - self.im_cell/2, ima + self.im_cell/2))

    @property
    def cell_limits(self):
        (rl, rr), (il, ir) = self.limits

        return (tuple(self.o.index.levels[0] - self.re_cell/2) + (rr,),
                tuple(self.o.index.levels[1] - self.im_cell/2) + (ir,))

    @property
    def bounds(self):
        return [[self.o.index.levels[0].min(), self.o.index.levels[0].max()],
                [self.o.index.levels[1].min(), self.o.index.levels[1].max()]]

    def scale_cell(self, k):
        self.re_cell *= k
        self.im_cell *= k

    def scale_index(self, k):
        self.o.index = self.o.index.map(lambda lab: (k*lab[0], k*lab[1]))

    def extend_index(self, copies, index_replicas):
        self.o = self.o.reindex(index=pandas.MultiIndex.from_product(
            [itertools.chain(*index_replicas[0]), itertools.chain(*index_replicas[1])],
            names=self.o.index.names))

    def replicas(self, copies):
        (rmi, rma), (imi, ima) = self.bounds
        rng = range(-copies, copies + 1)
        real = [numpy.around(
            self.o.index.levels[0] + m*(rma - rmi + self.re_cell), INDEX_ROUND_TO
        ) for m in rng]
        imag = [numpy.around(
            self.o.index.levels[1] + n*(ima - imi + self.im_cell), INDEX_ROUND_TO
        ) for n in rng]

        return real, imag

    def yield_vector_replicas(self, key, copies, index_replicas):
        current_vectors = self.o[key][~self.o[key].isna()]
        cases = itertools.product(enumerate(current_vectors), *index_replicas)

        for (k, vec), xrepl, yrepl in cases:
            index = pandas.MultiIndex.from_product([xrepl, yrepl])
            series = pandas.Series(vec.image.i.to_numpy(), index=index)

            yield complex(*index[k]), series

    def base_plot(self, axes, title=None):
        axes.set_title(title or str(self))
        xticks, yticks = self.cell_limits
        axes.xaxis.set_ticks(xticks, minor=True)
        axes.yaxis.set_ticks(yticks, minor=True)
        (rl, rr), (il, ir) = self.limits
        axes.set_xlim(rl, rr)
        axes.set_ylim(il, ir)
        axes.set_aspect('equal')
        axes.xaxis.set_label_text('Re')
        axes.yaxis.set_label_text('Im')

        return axes

    @stock.PyplotShow(figsize=(8, 8))
    def plot_domain(self, **kwargs):
        axes = self.base_plot(kwargs.pop('axes'))
        axes.grid(which='minor', color='w')
        axes.patch.set_facecolor('green')

    @stock.PyplotShow(figsize=(8, 8))
    def density_plot(self, vector, **kwargs):
        title = repr(vector)
        axes = self.base_plot(
            kwargs.pop('axes'), (title[:120] + '...') if len(title) > 120 else title)
        values = vector.image.density.unstack(level=0)
        xticks, yticks = self.cell_limits
        coll = axes.pcolormesh(xticks, yticks, values.values, **kwargs)
        kwargs.pop('figure').colorbar(coll, ax=axes)

    @classmethod
    def rectangle(cls, space_type, sw, ne, re_cells, im_cells=None):
        reals, re_cell = numpy.linspace(sw.real, ne.real, re_cells, retstep=True)
        imags, im_cell = numpy.linspace(sw.imag, ne.imag, im_cells or re_cells, retstep=True)
        index = pandas.MultiIndex.from_product([
            numpy.around(reals, INDEX_ROUND_TO), numpy.around(imags, INDEX_ROUND_TO)
        ], names=('real', 'imag'))

        return space_type(cls(re_cell, im_cell, index=index))



import abc

import pandas
import numpy

from hilbert import EQ_ROUND_TO

from hilbert import algebra
from hilbert import spaces
from hilbert import stock

from hilbert.curves import lib

EXPONENTS = {0: '', 1: '', 2: '²', 3: '³', 4: '⁴', 5: '⁵', 6: '⁶', 7: '⁷', 8: '⁸', 9: '⁹'}


class Image(stock.Repr, metaclass=abc.ABCMeta):
    def __init__(self, *args, series=None, **kwargs):
        self.i = pandas.Series(*args, **kwargs) if series is None else series

    def __str__(self):
        return f'\n{self.i}'

    @abc.abstractmethod
    def __getitem__(self, x):
        """Get image value at `x`"""

    @property
    def real(self):
        return pandas.Series(self.i.real, index=self.i.index)

    @property
    def imag(self):
        return pandas.Series(self.i.imag, index=self.i.index)

    @property
    def density(self):
        return self.i.abs()**2


class RImage(Image):
    def __getitem__(self, x):
        return self.i.loc[x]


class CImage(Image):
    def __getitem__(self, z):
        return self.i.loc[(z.real, z.imag)]


@stock.FrozenLazyAttrs(('space', 'curves', 'image_type'), ('image',))
class Vector(stock.Repr, stock.Eq, algebra.Vector):
    def __init__(self, space, *curves):
        self.space, self.curves = space, curves
        self.image_type = CImage if isinstance(self.space.bases, spaces.C1Field) else RImage

    def __call__(self, x):
        return self.image[x]

    def __str__(self):
        return ' + '.join(list(map(str, self.curves)))

    def _make_image(self):
        return self.image_type(series=sum(map(self.get_image_series, self.curves)))

    def get_image_series(self, curve):
        if isinstance(curve, lib.ImageCurve):
            return curve.image.i

        return self.image_type(
            curve(self.space.bases.domain()), index=self.space.bases.o.index).i

    def eq(self, other):
        return self.space == other.space and not (
            self.image.i - other.image.i).abs().round(EQ_ROUND_TO).any()

    def update(self):
        """Called to propagate space mutations when needed"""
        image = list(filter(lambda c: isinstance(c, lib.ImageCurve), self.curves))

        for curve in image:
            curve.image.i = curve.image.i.reindex(
                index=self.space.bases.o.index, fill_value=0)

        if self._frozen['image']:
            self.image.i = self.image.i.reindex(index=self.space.bases.o.index)
            nans = self.image.i[self.image.i.isna()]
            analytic = list(filter(
                lambda c: not isinstance(c, lib.ImageCurve), self.curves))
            values = pandas.Series(sum([
                c(self.space.bases.index_domain(nans.index)) for c in analytic
            ]) if analytic else [0]*len(nans.index), index=nans.index)
            self.image.i.update(values)

        return self

    def full_plot(self, top_ax, bottom_ax, **kwargs):
        self.image.real.plot(ax=top_ax, **kwargs)
        self.image.imag.plot(ax=bottom_ax, **kwargs)

    def density_plot(self, axes, **kwargs):
        self.image.density.plot(ax=axes, **kwargs)

    @stock.PyplotShow(nrows=2, ncols=1)
    def show(self, **kwargs):
        top_ax, bottom_ax = kwargs.pop('axes')
        self.full_plot(top_ax, bottom_ax, **kwargs)
        top_ax.yaxis.set_label_text('Re')
        bottom_ax.yaxis.set_label_text('Im')
        top_ax.grid()
        bottom_ax.grid()

    @stock.PyplotShow()
    def show_density(self, **kwargs):
        axes = kwargs.pop('axes')
        self.density_plot(axes, **kwargs)
        axes.yaxis.set_label_text('abs²')
        axes.grid()

    def kind(self):
        return '+'.join(sorted(curve.kind() for curve in self.curves))

    def eat(self, other):
        if isinstance(other, (int, float)):
            return self.__class__(self.space, lib.Polynomial(other))

        return other

    def num_prod(self, number):
        return self.__class__(self.space, *(number*cu for cu in self.curves))

    def add_other(self, other):
        self.check_if_equal_spaces(other)

        return self.__class__(self.space, *(self.curves + other.curves))

    def braket(self, other):
        self.check_if_equal_spaces(other)

        return self.space.bases.measure*numpy.dot(numpy.conj(self.image.i), other.image.i)

    def check_if_equal_spaces(self, other):
        if not self.space == other.space:
            raise NotImplementedError('Vectors should belong to the same space')
